# -*- coding: utf-8 -*-
"""
Created on Mon Apr 22 11:50:43 2024

@author: Jessica
"""

# %% Imports
import numpy as np
from enum import Enum
from Side import SideType

# %% Enum Category
class Category(Enum):
    """
    An enumeration representing different types of categories of a puzzle piece
    
    The possible types are:
        - EDGE: An edge piece.
        - CORNER: A corner piece.
        - INNER: An inner piece.
        
    """
    
    EDGE = 1
    CORNER = 2
    INNER = 3
    
# %% Class PuzzlePiece
class PuzzlePiece:
    """
    A class used to represent a puzzle piece.
    
    Attributes
    ----------
    __id : int
        id of the puzzle piece
    __category : Category
        a value of the enumeration Category
    __sides : np.array
        an array that contains the four sides of a puzzle piece
    __number_of_holes : int
        number of holes in the puzzle piece
    __number_of_heads : int 
        number of heads in the puzzle piece
    __top_side : int
        top side of the puzzle piece. The default is 0.
    __is_sorted : boolean
        decides if the piece is already sorted. The default is False.
    __image_path : string
        url / path of the image
    
    Methods
    -------
    rotate(top_side : int) : None

    """
        
    def __init__(self, id, category: Category, number_of_holes, number_of_heads, image_path, sides = np.array):
        """
        Initialises an instance of Puzzle.

        Parameters
        ----------
        category : Category
            a value of the enumeration Category
        number_of_holes : int
            number of holes in the puzzle piece
        number_of_heads : int
            number of heads in the puzzle piece
        image_path : string
            url / path of the image
        sides : np.array, optional
            an array that contains the four sides of a puzzle piece. The default is np.ndarray(shape=(4), dtype=object)

        """
        self.__id = id
        self.__sides = sides
        self.__category = category
        self.__number_of_holes = number_of_holes 
        self.__number_of_heads = number_of_heads
        self.__top_side = 0
        self.__is_sorted = False
        self.__image_path = image_path
        
    def get_id(self):
        return self.__id
    
    def get_sides(self):
        return self.__sides
    
    def set_sides(self, sides):
        self.__sides = sides
    
    def get_category(self):
        return self.__category
    
    def set_category(self, category):
        self.__category = category
    
    def get_number_of_holes(self):
        return self.__number_of_holes
    
    def set_number_of_holes(self, number_of_holes):
        self.__number_of_holes = number_of_holes
    
    def get_number_of_heads(self):
        return self.__number_of_heads
    
    def set_number_of_heads(self, number_of_heads):
        self.__number_of_heads = number_of_heads
        
    def get_top_side(self):
        return self.__top_side
    
    def set_top_side(self, top_side):
        self.__top_side = top_side
        if (self.__top_side < 0):
            self.__top_side = 3
        elif (self.__top_side > 3):
            #self.__top_side %= 4 #= 0
            self.__top_side -= 4
    
    def get_is_sorted(self):
        return self.__is_sorted
        
    def set_is_sorted(self, is_sorted):
        self.__is_sorted = is_sorted
        
    def get_image_path(self):
        return self.__image_path
    
    def set_image_path(self, image_path):
        self.__image_path = image_path
        
    def rotate(self, top_side):
        """
        Rotates a puzzle piece by setting the value top_side of the puzzle piece.
        Rotation is always clockwise.
        
        If a value between 0 and 3 is passed, 
            this value is assigned to the value top_side of the puzzle piece.
        
        If the value -1 is passed,
            corners will be adjusted to the upper left corner and
            edges will be adjusted to the upper edge.
        
        If the value -2 is passed,
            corners will be adjusted to the upper right corner and
            edges will be adjusted to the right edge.
        
        If the value -3 is passed,
            corners will be adjusted to the bottom left corner and
            edges will be adjusted to the bottom edge.
            
        If the value -4 is passed,
            corners will be adjusted to the bottom right corner and
            edges will be adjusted to the left edge.
            
        Parameters
        ----------
        top_side : int
            value for rotation

        Returns
        -------
        None.

        """
        
        # set top_side to passed value
        if top_side >= 0 and top_side <= 3: 
            self.top_side = top_side
        
        # rotate until suitable
        elif self.category == Category.CORNER:      # just for corners
            flat_sides = [s.id for s in self.sides if s.side_type == SideType.FLAT]
            min_id = min(flat_sides)
            max_id = max(flat_sides)
            
            if max_id == 3 and min_id == 0:
                min_id = 3
                max_id = 0
                
            #temp_top_side = self.top_side
            
            if top_side == -1: # upper left corner
                self.top_side = max_id #+= abs(max_id - temp_top_side)
            
            if top_side == -2: # upper right corner
                self.top_side = min_id #+= abs(min_id - temp_top_side)
            
            if top_side == -3: # bottom left corner
                self.top_side = max_id + 1 #+= (abs(max_id - temp_top_side) + 1)
            
            if top_side == -4: # bottom right corner
                self.top_side = min_id - 1 #+= abs(max_id - temp_top_side)
        
        elif self.category == Category.EDGE:      # just for edges
            flat_side = [s.id for s in self.sides if s.side_type == SideType.FLAT][0]
            self.top_side = flat_side
            
            if top_side == -1: # upper edge
                # nothing else to do
                pass
            
            if top_side == -2: # right edge
                self.top_side += 3
                
            if top_side == -3: # bottom edge
                self.top_side += 2
            
            if top_side == -4: # left edge
                self.top_side += 1        
    
    id = property(get_id)
    sides = property(get_sides, set_sides)
    category = property(get_category, set_category)
    number_of_holes = property(get_number_of_holes, set_number_of_holes)
    number_of_heads = property(get_number_of_heads, set_number_of_heads)
    top_side = property(get_top_side, set_top_side)
    is_sorted = property(get_is_sorted, set_is_sorted)
    image_path = property(get_image_path, set_image_path)