"""
Created on Fri Apr 26 15:15:23 2024

@author: Jan Genders
"""

import sys
import cv2  # Importiere OpenCV fuer Bildbearbeitung
import numpy as np  # Importiere NumPy fuer Array-Manipulation

sys.path.append('../general') # insert another input path

from PIL import Image, ImageDraw  # Importiere die Bildverarbeitungsbibliothek PIL
from PuzzlePiece import PuzzlePiece, Category  # Importiere PuzzlePiece und Category aus dem PuzzlePiece-Modul
from Puzzle import Puzzle  # Importiere Puzzle aus dem Puzzle-Modul
import matplotlib.pyplot as plt  # Importiere Matplotlib fuer die Visualisierung
import matplotlib.animation as animation  # Importiere Matplotlib fuer Animationen

class PuzzleConnector:
    """
    Klasse zum Verbinden von Puzzlestuecken zu einem vollstaendigen Bild.
    
    Attributes:
    - rows: Anzahl der Zeilen im Puzzle-Raster
    - columns: Anzahl der Spalten im Puzzle-Raster
    - puzzle_pieces: Liste der Puzzlestuecke als NumPy-Arrays
    """

    def __init__(self, rows, columns):
        """
        Initialisiert ein PuzzleConnector-Objekt.
        
        Eingaben:
        - rows: Anzahl der Zeilen des Puzzles
        - columns: Anzahl der Spalten des Puzzles
        """
        self.rows = rows
        self.columns = columns
        self.puzzle_pieces = []  # Leere Liste zur Speicherung der Puzzlestuecke

    def add_piece(self, piece):
        """
        Fügt ein Puzzlestück zur Liste der Puzzlestuecke hinzu.
        
        Eingaben:
        - piece: Ein NumPy-Array, das ein Puzzlestueck repräsentiert
        
        Raises:
        - ValueError: Wenn das Stueck kein gueltiges NumPy-Array ist oder eine falsche Form hat
        """
        if not isinstance(piece, np.ndarray):
            raise ValueError("The puzzle piece is not a valid numpy array.")
        if len(piece.shape) != 3 or piece.shape[2] != 3:
            raise ValueError("The puzzle piece does not have the correct shape. Expected (height, width, 3).")
        self.puzzle_pieces.append(piece)

    def compose(self):
        """
        Zusammensetzen der Puzzlestücke zu einem vollstuendigen Bild.
        
        Berechnet die Groesse des Endbilds basierend auf den Dimensionen der Puzzlestuecke
        und fuegt die Stuecke in der richtigen Reihenfolge zusammen.
        
        Rueckgabe:
        - composed_image: Das zusammengesetzte Bild als NumPy-Array
        """
        if len(self.puzzle_pieces) != self.rows * self.columns:
            raise ValueError("Number of puzzle pieces does not match the specified grid dimensions.")
        
        # Bestimme die maximale Breite und die Gesamthoehe fuer jede Zeile und Spalte
        max_widths = [0] * self.columns
        total_heights = [0] * self.rows

        piece_dimensions = []
        
        for index, piece in enumerate(self.puzzle_pieces):
            piece_height, piece_width, _ = piece.shape
            row = index // self.columns
            col = index % self.columns

            if col < len(max_widths):
                max_widths[col] = max(max_widths[col], piece_width)
            if row < len(total_heights):
                total_heights[row] = max(total_heights[row], piece_height)

            piece_dimensions.append((piece_height, piece_width))

        # Berechne die Gesamtbreite und -hoehe des zusammengesetzten Bildes
        total_width = sum(max_widths)
        total_height = sum(total_heights)

        composed_image = np.zeros((total_height, total_width, 3), dtype=np.uint8)
        
        current_y = 0
        for row in range(self.rows):
            current_x = 0
            for col in range(self.columns):
                index = row * self.columns + col
                piece = self.puzzle_pieces[index]
                piece_height, piece_width, _ = piece.shape

                # Fuege das Puzzlestueck in das zusammengesetzte Bild ein
                composed_image[current_y:current_y + piece_height, current_x:current_x + piece_width, :] = piece
                current_x += piece_width

            current_y += total_heights[row]

        return composed_image  # Rueckgabe des zusammengesetzten Bildes

    def rotate_pieces(self, puzzle):
        """
        Rotiert die Puzzlestuecke entsprechend ihrer Ausrichtung.
        
        Liest die Puzzlestuecke von ihrem Pfad, rotiert sie gemaess� ihrer
        Top-Side-Eigenschaft und fuegt sie dann der Liste der Puzzlestuecke hinzu.
        """
        for piece in puzzle.pieces:
            # Lade das Bild des Puzzlestuecks
            p = cv2.imread(piece.image_path)
            
            # Rotieren, wenn erforderlich
            if (piece.top_side == 0):
                pass  # Keine Rotation
            elif (piece.top_side == 1):
                p = cv2.rotate(p, cv2.ROTATE_90_COUNTERCLOCKWISE)  # 270 Grad Rotation
            elif (piece.top_side == 2):
                p = cv2.rotate(p, cv2.ROTATE_180)  # 180 Grad Rotation
            elif (piece.top_side == 3):
                p = cv2.rotate(p, cv2.ROTATE_90_CLOCKWISE)  # 90 Grad Rotation
            self.add_piece(p)  # Fuege das rotierte Stueck hinzu

class PuzzleConnectorWithAnimation(PuzzleConnector):
    """
    Erweiterung der PuzzleConnector-Klasse, um Animationen zu unterstützen.
    
    Attributes:
    - frames: Liste von Frames fuer die Animation, jeweils ein Bild
    """

    def __init__(self, rows, columns):
        """
        Initialisiert ein PuzzleConnectorWithAnimation-Objekt.
        
        Eingaben:
        - rows: Anzahl der Zeilen des Puzzles
        - columns: Anzahl der Spalten des Puzzles
        """
        super().__init__(rows, columns)
        self.frames = []  # Leere Liste zur Speicherung der Animationsframes

    def compose_with_animation(self):
        """
        Zusammensetzen der Puzzlestücke zu einem Bild und Speichern von Animationsframes.
        
        Die Funktion erstellt ein Bild mit jedem hinzugefuegten Puzzlestueck und speichert
        jedes Zwischenbild als Frame in der Animationsliste.
        
        Rueckgabe:
        - None: Fuegt Frames der Animationsliste hinzu und speichert ein Beispielbild
        """
        if len(self.puzzle_pieces) != self.rows * self.columns:
            raise ValueError("Number of puzzle pieces does not match the specified grid dimensions.")
        
        max_widths = [0] * self.columns
        total_heights = [0] * self.rows
        piece_dimensions = []
        
        for index, piece in enumerate(self.puzzle_pieces):
            piece_height, piece_width, _ = piece.shape
            row = index // self.columns
            col = index % self.columns

            if col < len(max_widths):
                max_widths[col] = max(max_widths[col], piece_width)
            if row < len(total_heights):
                total_heights[row] = max(total_heights[row], piece_height)

            piece_dimensions.append((piece_height, piece_width))

        total_width = sum(max_widths)
        total_height = sum(total_heights)

        composed_image = np.zeros((total_height, total_width, 3), dtype=np.uint8)
        
        current_y = 0
        for row in range(self.rows):
            current_x = 0
            for col in range(self.columns):
                index = row * self.columns + col
                piece = self.puzzle_pieces[index]
                piece_height, piece_width, _ = piece.shape

                composed_image[current_y:current_y + piece_height, current_x:current_x + piece_width, :] = piece
                current_x += piece_width

                # Fuege das aktuelle Bild als Frame zur Animation hinzu
                frame = Image.fromarray(cv2.cvtColor(composed_image, cv2.COLOR_BGR2RGB))
                self.frames.append(frame)

            current_y += total_heights[row]

        print(f"Number of frames: {len(self.frames)}")  # Debug-Ausgabe der Anzahl der Frames
        if self.frames:
            self.frames[0].save('sample_frame.jpg')  # Speichern eines Beispielframes zur Überpruefung
            print("Sample frame saved as 'sample_frame.jpg'")

    def animate_puzzle(self):
        """
        Erstellt und zeigt die Animation des Puzzles.
        
        Ruft compose_with_animation() auf, um die Frames zu erstellen, und
        verwendet dann Matplotlib, um eine Animation zu erstellen und anzuzeigen.
        
        Rueckgabe:
        - None: Zeigt die Animation an und speichert sie als GIF
        """
        self.compose_with_animation()
        
        if not self.frames:
            print("No frames to animate!")  # Debug-Ausgabe, wenn keine Frames vorhanden sind
            return
        
        fig, ax = plt.subplots()  # Erstelle eine Matplotlib-Figur und Achse
        plt.axis('on')  # Schalte Achsen ein fuer Debugging
        
        def update_frame(num):
            """
            Aktualisiert den Frame der Animation.
            
            Eingaben:
            - num: Index des aktuellen Frames
            
            Rueckgabe:
            - None: Setzt das Bild des aktuellen Frames auf die Achse
            """
            ax.clear()  # Loesche die aktuelle Achse
            ax.imshow(self.frames[num])  # Zeige das Bild des aktuellen Frames
            plt.axis('on')  # Schalte Achsen ein fuer Debugging

        # Erstelle die Animation
        anim = animation.FuncAnimation(fig, update_frame, frames=len(self.frames), interval=500, repeat=False)
        
        # Speichere die Animation als GIF
        anim.save('puzzle_animation.gif', writer='imagemagick')
        print("Animation saved as 'puzzle_animation.gif'")

        plt.show()  # Zeige die Animation an

# Beispielverwendung
if __name__ == "__main__":
    # Initialisiere ein Puzzle-Objekt mit den Abmessungen 3x3 -- DEMO!
    solved_puzzle = Puzzle(3 ,3)
    
    # Erstelle Testdaten fuer die Puzzlestuecke
    piece0 = PuzzlePiece(0, Category.CORNER, 0, 2, '../../resources/demo_images/piece_1.png', None)
    piece0.top_side = 2
    piece1 = PuzzlePiece(1, Category.EDGE, 1, 2, '../../resources/demo_images/piece_2.png', None)
    piece1.top_side = 3
    piece2 = PuzzlePiece(2, Category.CORNER, 2, 0, '../../resources/demo_images/piece_3.png', None)
    piece2.top_side = 1
    piece3 = PuzzlePiece(3, Category.EDGE, 2, 1, '../../resources/demo_images/piece_4.png', None)
    piece4 = PuzzlePiece(4, Category.INNER, 2, 2, '../../resources/demo_images/piece_5.png', None)
    piece5 = PuzzlePiece(5, Category.EDGE, 2, 1, '../../resources/demo_images/piece_6.png', None)
    piece6 = PuzzlePiece(6, Category.CORNER, 2, 0, '../../resources/demo_images/piece_7.png', None)
    piece7 = PuzzlePiece(7, Category.EDGE, 1, 2, '../../resources/demo_images/piece_8.png', None)
    piece8 = PuzzlePiece(8, Category.CORNER, 0, 2, '../../resources/demo_images/piece_9.png', None)
    solved_puzzle.set_pieces(np.array([piece0, piece1, piece2, piece3, piece4, piece5, piece6, piece7, piece8]))
    
    # Definiere die Dimensionen des Rasters
    rows = solved_puzzle.height
    columns = solved_puzzle.width

    # Erstelle ein PuzzleConnectorWithAnimation-Objekt
    puzzle = PuzzleConnectorWithAnimation(rows, columns)
    puzzle.rotate_pieces(solved_puzzle)  # Drehe die Puzzlestuecke gemaess ihrer Ausrichtung

    # Animationsprozess starten
    puzzle.animate_puzzle()
