"""
Created on Fri Apr 26 15:09:52 2024

@author: Jan Genders
"""

import cv2
import numpy as np
from scipy import ndimage

class PuzzlePieceExtractor:
    def __init__(self, threshold=50, minimum_res=4000):
        """
        Initialisiert den PuzzlePieceExtractor mit einem Schwellenwert fuer die Farbtoleranz 
        und einer Mindestaufloesung fuer die Puzzleteile.
        """
        self.threshold = threshold
        self.minimum_res = minimum_res
        self.clicked_point_background = None
        self.clicked_point_shadow = None
    
    # ------------------------------------------
    #   Event Handler: Farbe durch Klicken auswaehlen
    # ------------------------------------------
    def get_color(self, event, x, y, flags, param):
        """
        Event-Handler zum Erfassen von Mausklicks, um die Positionen fuer 
        Hintergrund- und Schattenfarbe zu bestimmen.
        """
        if event == cv2.EVENT_LBUTTONDOWN:
            print(x,y)
            if self.clicked_point_background is None:
                self.clicked_point_background = (x,y)#(1472, 932) Example
            elif self.clicked_point_shadow is None:
                self.clicked_point_shadow = (x,y)#(42, 98) Example
    
    # ------------------------------------------
    #   Hintergrund- und Schattenfarbe auswaehlen
    # ------------------------------------------
    def select_colors(self, image):
        """
        Ermoeglicht dem Benutzer, durch Klicken die Hintergrund- und Schattenfarbe 
        im Bild auszuwaehlen.
        """
        clone = image.copy()
        cv2.namedWindow("Select Background and Shadow Color")
        cv2.setMouseCallback("Select Background and Shadow Color", self.get_color)

        while True:
            clone_draw = clone.copy()
            if self.clicked_point_background:
                cv2.circle(clone_draw, self.clicked_point_background, 5, (0, 255, 0), -1)
            if self.clicked_point_shadow:
                cv2.circle(clone_draw, self.clicked_point_shadow, 5, (0, 0, 255), -1)
            cv2.imshow("Select Background and Shadow Color", clone_draw)

            key = cv2.waitKey(1) & 0xFF
            if key == ord("c"):  # 'c' druecken, um die Auswahl zu bestaetigen
                cv2.destroyAllWindows()
                break

        # Farben an den ausgewaehlten Punkten berechnen
        x_bg, y_bg = self.clicked_point_background
        x_sh, y_sh = self.clicked_point_shadow

        background_color = (image[y_bg, x_bg]/9 + image[y_bg+1, x_bg]/9 + image[y_bg+1, x_bg+1]/9 
                            + image[y_bg+1, x_bg-1]/9 + image[y_bg-1, x_bg]/9 + image[y_bg-1, x_bg-1]/9 
                            + image[y_bg-1, x_bg+1]/9 + image[y_bg, x_bg-1]/9 + image[y_bg, x_bg+1]/9)

        shadow_color = (image[y_sh, x_sh]/9 + image[y_sh+1, x_sh]/9 + image[y_sh+1, x_sh+1]/9 
                        + image[y_sh+1, x_sh-1]/9 + image[y_sh-1, x_sh]/9 + image[y_sh-1, x_sh-1]/9 
                        + image[y_sh-1, x_sh+1]/9 + image[y_sh, x_sh-1]/9 + image[y_sh, x_sh+1]/9)
        
        return background_color, shadow_color
    
    # ------------------------------------------
    #   Hintergrund und Schatten entfernen
    # ------------------------------------------
    def remove_background(self, image, background_color, shadow_color):
        """
        Entfernt den Hintergrund und den Schatten basierend auf den ausgewaehlten Farben.
        """
        image_blur = cv2.blur(image, (7, 7))  # Bild leicht unscharf machen, um Stoerungen zu reduzieren
        hundred = np.array([255, 255, 255])
        th_bg = (background_color / hundred) * self.threshold  # Schwellenwert fuer Hintergrundfarbe
        th_sh = (shadow_color / hundred) * self.threshold  # Schwellenwert fuer Schattenfarbe

        # Farbgrenzen fuer Hintergrund und Schatten definieren
        lower_bound_bg = background_color - th_bg
        upper_bound_bg = background_color + th_bg

        lower_bound_sh = shadow_color - th_sh
        upper_bound_sh = shadow_color + th_sh

        # Masken fuer Hintergrund und Schatten erstellen
        mask_bg = cv2.inRange(image_blur, lower_bound_bg, upper_bound_bg)
        mask_sh = cv2.inRange(image_blur, lower_bound_sh, upper_bound_sh)

        # Masken kombinieren und auf das Bild anwenden
        combined_mask = cv2.bitwise_or(mask_bg, mask_sh)
        result = cv2.bitwise_and(image, image, mask=~combined_mask)
        return result
    
    # ------------------------------------------
    #   Bild erodieren und dilatieren
    # ------------------------------------------
    def erode_n_dilate(self, image):
        """
        Wendet eine Erosion und Dilatation auf das Bild an, um kleinere Artefakte zu entfernen.
        """
        kernel = np.ones((3, 3), np.uint8)  # Kernel fuer Morphologie-Operationen
        img_erode = cv2.erode(image, kernel, iterations=6)  # Erosion anwenden
        img_dilate = cv2.dilate(img_erode, kernel, iterations=6)  # Dilatation anwenden

        # Entfernen von Restartefakten nach Morphologie-Operationen
        length, height = img_dilate.shape[:2]
        for i in range(length):
            for j in range(height):
                if all(img_dilate[i, j] == 0):
                    image[i, j] = 0

        return image
    
    # ------------------------------------------
    #   Puzzle-Teile im Bild erkennen und extrahieren
    # ------------------------------------------
    def splitting(self, image):
        """
        Findet und extrahiert die Puzzle-Teile aus dem Bild.
        """
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)  # Bild in Graustufen umwandeln
        _, binary = cv2.threshold(gray, 1, 255, cv2.THRESH_BINARY)  # Binaerbild erstellen
        contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)  # Konturen finden

        puzzle_pieces = []
        for contour in contours:
            x, y, w, h = cv2.boundingRect(contour)  # Begrenzungsrechteck fuer jede Kontur
            if (h * w) > self.minimum_res:  # Puzzle-Teil nur speichern, wenn es gross genug ist
                piece = np.copy(image[(y-1):y+(h+2), (x-1):x+(w+2)])
                clean_piece = self.remove_border_artifacts(piece)  # Raender von Artefakten bereinigen
                puzzle_pieces.append(clean_piece)

        return puzzle_pieces
    
    # ------------------------------------------
    #   Artefakte am Rand entfernen
    # ------------------------------------------
    def remove_border_artifacts(self, image_array):
        """
        Entfernt unerwuenschte Randartefakte von extrahierten Puzzle-Teilen.
        """
        grayscale_image = cv2.cvtColor(image_array, cv2.COLOR_BGR2GRAY) if len(image_array.shape) == 3 else image_array
        _, binary_image = cv2.threshold(grayscale_image, 1, 255, cv2.THRESH_BINARY)
        num_labels, labels, stats, centroids = cv2.connectedComponentsWithStats(binary_image, connectivity=8)
        mask = np.ones_like(binary_image, dtype=np.uint8) * 255  # Maske initialisieren

        for i in range(1, num_labels):
            x, y, w, h, area = stats[i]
            # Komponenten, die den Rand beruehren, aus der Maske entfernen
            if x == 0 or y == 0 or x + w == binary_image.shape[1] or y + h == binary_image.shape[0]:
                mask[labels == i] = 0

        # Maske auf das Bild anwenden
        mask = cv2.merge([mask, mask, mask])
        cleaned_image = cv2.bitwise_and(image_array, mask)
        return cleaned_image
    
    # ------------------------------------------
    #   Hauptmethode zum Erzeugen der Puzzle-Teile
    # ------------------------------------------
    def generate_pieces(self, image):
        """
        Fuehrt den gesamten Prozess von der Farbauswahl bis zur Erzeugung 
        und Rueckgabe der Puzzle-Teile durch.
        """
        background_color, shadow_color = self.select_colors(image)
        result = self.remove_background(image, background_color, shadow_color)
        result_image = self.erode_n_dilate(result)
        pieces = self.splitting(result_image)

        # Konvertiert die Puzzleteile in Graustufen-Binaerbilder
        gray_pieces = []
        for piece in pieces:
            gray = cv2.cvtColor(piece, cv2.COLOR_BGR2GRAY)
            _, binary = cv2.threshold(gray, 1, 255, cv2.THRESH_BINARY)
            gray_pieces.append(binary)

        return pieces, gray_pieces


if __name__ == "__main__":
    # Bild laden und Puzzle-Teile extrahieren
    image = cv2.imread("../../resources/demo_images/all_mixed.jpg")
    extractor = PuzzlePieceExtractor()
    pieces, gray_pieces = extractor.generate_pieces(image)

    # Puzzleteile anzeigen
    for piece in pieces:
        cv2.imshow("Single Piece", piece)
        cv2.waitKey(0)
        cv2.destroyAllWindows()
