# -*- coding: utf-8 -*-
"""
Created on Fri Apr 26 15:19:51 2024

@author: Jessica
"""

import sys
import numpy as np

sys.path.append('../general') # insert another input path

from Puzzle import Puzzle
from PuzzlePiece import PuzzlePiece
from PuzzlePiece import Category
from Side import Side
from Side import SideType
from MatchingCounterpiece import MatchingCounterpiece

class PuzzleSolver:
    """
    A class used to represent the puzzle solver.
        
    Methods
    -------
    get_compatible_side_type(side_type : SideType) : SideType
    get_possible_sides(piece : PuzzlePiece, current_side : Side, idx_possible_side : int) : np.array
    find_compatible(current_piece : PuzzlePiece, current_side : Side, possible_pieces : np.array, idx_help : int) : (PuzzlePiece, Side)
    solve(puzzle : Puzzle) : Puzzle
    
    """
    
    @staticmethod
    def get_compatible_side_type(side_type: SideType):
        """
        Determines the compatible side type to a given side type.

        Parameters
        ----------
        side_type : SideType
            side_type for which the matching side type is to be found

        Returns
        -------
        SideType
            matching side type to given side type

        """
        if side_type == SideType.HEAD:
            return SideType.HOLE
        elif side_type == SideType.HOLE:
            return SideType.HEAD
        else:
            return None
    
    @staticmethod
    def get_possible_sides(piece: PuzzlePiece, current_side: Side, idx_possible_side):
        """
        Finds the possible sides of a given puzzle piece.

        Parameters
        ----------
        piece : PuzzlePiece
            puzzle piece for which the possible sides are to be found
        current_side : Side
            side of the current piece that is being considered
        idx_possible_side : int
            auxiliary index for determining the possible side(s)

        Returns
        -------
        possible_sides : np.array
            an array that contains the possible side(s)

        """
        possible_sides = np.array([])
        if piece.category == Category.INNER:
            for s in piece.sides:
                if ((s.side_type == PuzzleSolver.get_compatible_side_type(current_side.side_type)) and (s.height == current_side.height) and (s.width == current_side.width) and (s.is_sorted == False)):
                    possible_sides = np.append(possible_sides, s)
        else:
            idx_possible_side = piece.top_side + idx_possible_side
            if (idx_possible_side < 0):
                 idx_possible_side = 3
            elif (idx_possible_side > 3):
                 idx_possible_side %= 4
            s = piece.sides[idx_possible_side]
            if ((s.side_type == PuzzleSolver.get_compatible_side_type(current_side.side_type)) and (s.height == current_side.height) and (s.width == current_side.width) and (s.is_sorted == False)):
                possible_sides = np.append(possible_sides, s)
        
        return possible_sides  

    @staticmethod
    def find_compatible(current_piece: PuzzlePiece, current_side: Side, possible_pieces, idx_help = 0):
        """
        Finds the compatible piece to a given piece by calculating the distance of the outline of the sides. 
        The matching piece corresponds to the piece with the smallest distance.
        If one is found the value is_sorted of the piece and the value is_sorted of the side are set to "True".
        Returns the compatible piece and the compatible side.

        Parameters
        ----------
        current_piece : PuzzlePiece
            piece for which a compatible piece is to be found
        current_side : Side
            side of the current piece that is being considered
        possible_pieces : np.array
            an array that contains all possibly matching pieces
        idx_help : int
            auxiliary index for determining the possible side(s)

        Returns
        -------
        PuzzlePiece
            the compatible piece
        Side
            the compatible side

        """
        matching_counterpieces = np.array([])
        possible_sides = np.array([])
        for p in possible_pieces:
            possible_sides = np.array([])
            possible_sides = np.append(possible_sides, PuzzleSolver.get_possible_sides(p, current_side, idx_help))
            if possible_sides.size == 0:
                pass
            elif possible_sides.size == 1:
                distance = np.linalg.norm(possible_sides[0].outline - current_side.outline)
                matching_counterpieces = np.append(matching_counterpieces, MatchingCounterpiece(p, possible_sides[0], distance))
            else:
                distance = [np.linalg.norm(s.outline - current_side.outline) for s in possible_sides]
                s = possible_sides[np.argmin(distance)]
                matching_counterpieces = np.append(matching_counterpieces, MatchingCounterpiece(p, s, distance))
        
        min_distance_idx = np.argmin(np.array([m.distance for m in matching_counterpieces]))
        piece = matching_counterpieces[min_distance_idx].piece
        side = matching_counterpieces[min_distance_idx].side
        piece.is_sorted = True
        side.is_sorted = True
        print("compatible to current piece", current_piece.id, " (side ", current_side.id, "): piece", piece.id, "(side ", side.id, ")")
        return (piece, side)
    
    @staticmethod
    def solve(puzzle: Puzzle):
        """
        Solves a given puzzle by sorting the puzzle pieces.

        Parameters
        ----------
        puzzle : Puzzle
            puzzle to be sorted

        Returns
        -------
        sorted_puzzle : TYPE
            sorted puzzle

        """
        
        # create empty puzzle to store sorted puzzle pieces
        sorted_puzzle = Puzzle(puzzle.width, puzzle.height)
       
        # set start corner        
        # get_corners
        corners = puzzle.get_corners()
        # set start piece
        current_piece = corners[0]
        current_piece.rotate(-1)
        current_piece.is_sorted = True        
        # set start side
        current_side = current_piece.sides[current_piece.top_side + 1]
        current_side.is_sorted = True
        
        sorted_puzzle.set_piece(current_piece, 0)
        puzzle_is_sorted = False

        idx = 1
        row = 0
        height = 0
        width = 1

        # sort puzzle
        while puzzle_is_sorted == False:
                
            # maximum width - 1 reached -> search for last piece of the row
            if width == puzzle.width - 1:
                # first or last row -> search for corner
                # possible side of the searched corner 
                #   -> first row: rotate(-2), top_side - 1
                #   -> last row: rotate(-4), top_side - 1
                if height == 0 or height == puzzle.height - 1: 
                    corners = puzzle.get_corners()
                    if height == 0:
                        for c in corners: 
                            c.rotate(-2)
                    else:
                        for c in corners:
                            c.rotate(-4)
                    (current_piece, current_side) = PuzzleSolver.find_compatible(current_piece, current_side, corners, -1)
                    sorted_puzzle.set_piece(current_piece, idx)
                    # last row -> puzzle is sorted
                    if height == puzzle.height - 1:
                        puzzle_is_sorted = True
                # middle row -> search for edge piece
                # possible side of the searched edge piece -> rotate(-2), top_side - 1
                else:
                    edge_pieces = puzzle.get_edge_pieces()
                    for e in edge_pieces:
                        e.rotate(-2)
                    (current_piece, current_side) = PuzzleSolver.find_compatible(current_piece, current_side, edge_pieces, -1)
                    sorted_puzzle.set_piece(current_piece, idx)           
                idx += 1
                width += 1
                
            # maximum width reached -> search for first piece of the next row
            elif width == puzzle.width:
                # start new row
                row += 1
                width = 0
                height += 1
                # reset current_piece & current_side: first piece in row above, bottom side
                current_piece = sorted_puzzle.pieces[idx - puzzle.width]
                current_side = current_piece.sides[(current_piece.top_side + 2) % 4]
                # last row -> search for corner
                # possible side of the searched corner -> rotate(-3), top_side
                if row == puzzle.height - 1:
                    corners = puzzle.get_corners()
                    for c in corners: 
                        c.rotate(-3)
                    (current_piece, current_side) = PuzzleSolver.find_compatible(current_piece, current_side, corners, 0)  
                    sorted_puzzle.set_piece(current_piece, idx)
                    current_side = current_piece.sides[(current_piece.top_side + 1) % 4]
                # middle row -> search for edge piece
                # possible side of the searched edge piece -> rotate(-4)
                else:
                    edge_pieces = puzzle.get_edge_pieces()
                    for e in edge_pieces:
                        e.rotate(-4)
                    (current_piece, current_side) = PuzzleSolver.find_compatible(current_piece, current_side, edge_pieces, 0)
                    sorted_puzzle.set_piece(current_piece, idx)  
                current_side = current_piece.sides[(current_piece.top_side + 1) % 4]
                idx += 1
                width += 1
            
            # any other row -> search for middle piece (inner piece or edge piece)
            else: 
                # first or last row -> search for edge piece
                # possible side of the searched edge piece 
                #   -> first row: rotate(-1), top_side - 1
                #   -> last row: rotate(-3), top_side - 1
                if row == 0 or row == puzzle.height - 1:
                    edge_pieces = puzzle.get_edge_pieces()
                    if row == 0: # first row
                        for e in edge_pieces:
                            e.rotate(-1)
                    else: # last row
                        for e in edge_pieces:
                            e.rotate(-3)
                    (current_piece, current_side) = PuzzleSolver.find_compatible(current_piece, current_side, edge_pieces, -1)
                    sorted_puzzle.set_piece(current_piece, idx)
                    current_side = current_piece.sides[(current_piece.top_side + 1) % 4]
                # any other row -> search for inner piece
                # any side is possible
                else: # inner piece
                    inner_pieces = puzzle.get_inner_pieces()
                    (current_piece, current_side) = PuzzleSolver.find_compatible(current_piece, current_side, inner_pieces)
                    current_piece.rotate(current_side.id + 1)
                    sorted_puzzle.set_piece(current_piece, idx)
                    current_side = current_piece.sides[(current_piece.top_side + 1) % 4]
                idx += 1
                width += 1
    
        return sorted_puzzle

#%% test script
if __name__ == "__main__":
    # additional imports for test data
    sys.path.append('../../resources/test data')
    from testPuzzleData import get_puzzle_2x2
    from testPuzzleData import get_puzzle_3x3

    puzzle = get_puzzle_3x3()
    sorted_puzzle = PuzzleSolver.solve(puzzle)
    for p in sorted_puzzle.pieces:
        print("top side piece ", p.id, " = ", p.top_side)

